from pathlib import Path
from typing import List, Optional

from otlmow_modelbuilder.NewOTLBaseClassNotImplemented import NewOTLBaseClassNotImplemented
from otlmow_modelbuilder.SQLDataClasses.GeneralInfoRecord import GeneralInfoRecord
from otlmow_modelbuilder.SQLDataClasses.Inheritance import Inheritance
from otlmow_modelbuilder.SQLDataClasses.OSLOAttribuut import OSLOAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeComplexAttribuut import OSLODatatypeComplexAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypePrimitiveAttribuut import OSLODatatypePrimitiveAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeUnion import OSLODatatypeUnion
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeUnionAttribuut import OSLODatatypeUnionAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeComplex import OSLODatatypeComplex
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypePrimitive import OSLODatatypePrimitive
from otlmow_modelbuilder.SQLDataClasses.OSLOEnumeration import OSLOEnumeration
from otlmow_modelbuilder.OSLOInMemoryCreator import OSLOInMemoryCreator
from otlmow_modelbuilder.SQLDataClasses.OSLORelatie import OSLORelatie
from otlmow_modelbuilder.SQLDataClasses.OSLOTypeLink import OSLOTypeLink


class OSLOCollector:
    def __init__(self, path: Path):
        self.path = path

        self.inheritances: [Inheritance] = None
        self.attributes: [OSLOAttribuut] = None
        self.classes: [OSLOClass] = None
        self.primitive_datatypes: [OSLODatatypePrimitive] = None
        self.primitive_datatype_attributen: [OSLODatatypePrimitiveAttribuut] = None
        self.complex_datatypes: [OSLODatatypeComplex] = None
        self.complex_datatype_attributen: [OSLODatatypeComplexAttribuut] = None
        self.union_datatypes: [OSLODatatypeUnion] = None
        self.union_datatype_attributen: [OSLODatatypeUnionAttribuut] = None
        self.enumerations: [OSLOEnumeration] = None
        self.typeLinks: [OSLOTypeLink] = None
        self.relations: [OSLORelatie] = None
        self.general_info: [GeneralInfoRecord] = None

        self.class_dict: Optional[dict] = None

    def collect_all(self, include_abstract: bool = False) -> None:
        self.class_dict = {}
        with OSLOInMemoryCreator(self.path) as memory_creator:
            self.classes, self.class_dict = memory_creator.get_all_classes_and_class_dict()
            self.attributes = memory_creator.get_all_attributes(include_abstract=include_abstract)
            self.inheritances = memory_creator.get_all_inheritances()
            self.primitive_datatypes = memory_creator.get_all_primitive_datatypes()
            self.primitive_datatype_attributen = memory_creator.get_all_primitive_datatype_attributes()
            self.complex_datatypes = memory_creator.get_all_complex_datatypes()
            self.complex_datatype_attributen = memory_creator.get_all_complex_datatype_attributes()
            self.union_datatypes = memory_creator.get_all_union_datatypes()
            self.union_datatype_attributen = memory_creator.get_all_union_datatype_attributes()
            self.enumerations = memory_creator.get_all_enumerations()
            self.typeLinks = memory_creator.get_all_typelinks()
            self.relations = memory_creator.get_all_relations()
            self.general_info = memory_creator.get_general_info()

    def find_attributes_by_class(self, oslo_class: OSLOClass) -> [OSLOAttribuut]:
        if oslo_class is None:
            return []

        return sorted((a for a in self.attributes if a.class_uri == oslo_class.objectUri),
                      key=lambda a: a.objectUri)

    def find_inheritances_by_class(self, oslo_class: OSLOClass) -> [Inheritance]:
        if oslo_class is None:
            return []
        return self.find_inheritances_by_class_uri(oslo_class.objectUri)

    def find_inheritances_by_class_uri(self, oslo_class_uri: str) -> [Inheritance]:
        if oslo_class_uri is None or oslo_class_uri == '':
            return []

        return sorted((i for i in self.inheritances if i.class_uri == oslo_class_uri),
                      key=lambda i: i.base_uri)

    def find_all_indirect_inheritances_by_class_uri(self, oslo_class_uri: str) -> [Inheritance]:
        if oslo_class_uri is None or oslo_class_uri == '':
            return []

        inheritance_list = []
        direct_inheritances = self.find_inheritances_by_class_uri(oslo_class_uri)
        if len(direct_inheritances) == 0:
            return inheritance_list

        inheritance_list.extend(direct_inheritances)
        for direct in direct_inheritances:
            inheritance_list.extend(self.find_all_indirect_inheritances_by_class_uri(direct.class_uri))

        return sorted(inheritance_list, key=lambda c: c.base_uri)

    def find_subclasses_uri_by_class_uri(self, oslo_class_uri: str) -> [str]:
        if oslo_class_uri is None or oslo_class_uri == '':
            return []

        return sorted((i.class_uri for i in self.inheritances if i.base_uri == oslo_class_uri))

    def find_indirect_subclasses_uri_by_class_uri(self, oslo_class_uri: str) -> [str]:
        if oslo_class_uri is None or oslo_class_uri == '':
            return []

        superclass_list = []
        direct_superclasses = self.find_subclasses_uri_by_class_uri(oslo_class_uri)
        if len(direct_superclasses) == 0:
            return superclass_list

        superclass_list.extend(direct_superclasses)
        for direct in direct_superclasses:
            superclass_list.extend(self.find_indirect_subclasses_uri_by_class_uri(direct))

        return sorted(superclass_list)

    def find_class_by_uri(self, uri: str) -> OSLOClass:
        return next((p for p in self.classes if p.objectUri == uri), None)

    def find_primitive_datatype_by_uri(self, uri: str) -> OSLODatatypePrimitive:
        return next((p for p in self.primitive_datatypes if p.objectUri == uri), None)

    def find_primitive_datatype_attributes_by_class_uri(self, class_uri: str) -> List[OSLODatatypePrimitiveAttribuut]:
        return sorted((p for p in self.primitive_datatype_attributen if p.class_uri == class_uri),
                      key=lambda p: p.objectUri)

    def find_complex_datatype_by_uri(self, uri) -> OSLODatatypeComplex:
        return next((p for p in self.complex_datatypes if p.objectUri == uri), None)

    def find_complex_datatype_attributes_by_class_uri(self, class_uri: str) -> List[OSLODatatypeComplexAttribuut]:
        return sorted((p for p in self.complex_datatype_attributen if p.class_uri == class_uri),
                      key=lambda p: p.objectUri)

    def find_union_datatype_by_uri(self, uri) -> OSLODatatypeUnion:
        return next((p for p in self.union_datatypes if p.objectUri == uri), None)

    def find_union_datatype_attributes_by_class_uri(self, class_uri: str) -> List[OSLODatatypeUnionAttribuut]:
        return sorted((p for p in self.union_datatype_attributen if p.class_uri == class_uri),
                      key=lambda p: p.objectUri)

    def find_enumeration_by_uri(self, uri) -> OSLOEnumeration:
        return next((p for p in self.enumerations if p.objectUri == uri), None)

    def find_enumeration_by_codelist(self, uri) -> OSLOEnumeration:
        return next((p for p in self.enumerations if p.codelist == uri), None)

    def find_type_link_by_uri(self, type_uri: str) -> OSLOTypeLink:
        return next((p for p in self.typeLinks if p.item_uri == type_uri), None)

    def find_outgoing_relations(self, objectUri: str) -> [OSLORelatie]:
        return sorted((r for r in self.relations if r.bron_uri == objectUri and r.bron_overerving == ''
                       and r.doel_overerving == ''), key=lambda r: r.objectUri)

    def find_all_relations(self, objectUri: str, allow_duplicates: bool = False) -> [OSLORelatie]:
        """finds all relations, given an objectUri, where the object is either the source or the target of the relation.
        allow_duplicates is relevant for unidirectional relations, as the relation would be included twice:
        once where objectUri is the source and once where objectUri is the target. If allow_duplicates is False,
        only the relation where objectUri is the source is returned."""
        all_relations = [r for r in self.relations if (r.bron_uri == objectUri or r.doel_uri == objectUri)
                         and r.bron_overerving == '' and r.doel_overerving == '']
        if allow_duplicates:
            return sorted(all_relations, key=lambda r: r.objectUri)
        else:
            return sorted((r for r in all_relations
                           if r.richting == 'Unspecified' and r.bron_uri == objectUri or r.richting != 'Unspecified'),
                          key=lambda r: r.objectUri)

    def query_correct_base_classes(self) -> None:
        with OSLOInMemoryCreator(self.path) as memory_creator:
            result_uris = memory_creator.check_on_base_classes()
            if len(result_uris) != 0:
                print('The following classes are not using the correct base classes:')
                print(result_uris)
                raise NewOTLBaseClassNotImplemented()

    def find_all_concrete_relations(self, objectUri: str, allow_duplicates: bool = False) -> [OSLORelatie]:
        class_ = self.find_class_by_uri(objectUri)
        if class_ is None:
            raise ValueError(f'Class with uri {objectUri} does not exist.')
        if class_.abstract == 1:
            raise ValueError(f'Class with uri {objectUri} is an abstract class.')

        all_relations = [r for r in self.relations if (r.bron_uri == objectUri or r.doel_uri == objectUri)
                         and self.class_dict[r.bron_uri].abstract == 0 and self.class_dict[r.doel_uri].abstract == 0]

        if allow_duplicates:
            return sorted(all_relations, key=lambda r: r.objectUri)
        else:
            return sorted((r for r in all_relations
                           if (r.richting == 'Unspecified' and
                               (r.bron_uri == objectUri or r.bron_overerving == objectUri))
                           or r.richting != 'Unspecified'),
                           key=lambda r: r.objectUri)



import logging
import re
from typing import Dict

from otlmow_modelbuilder.AbstractDatatypeCreator import AbstractDatatypeCreator
from otlmow_modelbuilder.OSLOCollector import OSLOCollector
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeUnion import OSLODatatypeUnion


class OTLUnionDatatypeCreator(AbstractDatatypeCreator):
    def __init__(self, oslo_collector: OSLOCollector):
        super().__init__(oslo_collector)
        logging.info("Created an instance of OTLUnionDatatypeCreator")

    def create_block_to_write_from_union_types(self, union_datatype: OSLODatatypeUnion,
                                               union_datatype_validation_rules: Dict, model_location: str = '') -> [str]:
        if not isinstance(union_datatype, OSLODatatypeUnion):
            raise ValueError(f"Input is not a OSLODatatypeUnion")

        if union_datatype.objectUri == '':
            raise ValueError(f"OSLODatatypeUnion.objectUri is invalid. Value = '{union_datatype.objectUri}'")

        if union_datatype.objectUri in union_datatype_validation_rules['valid_uri_and_types'].keys():
            pass
        else:
            match_re = False
            for regex in union_datatype_validation_rules["valid_regexes"]:
                match_re = re.match(pattern=regex, string=union_datatype.objectUri)
                if match_re:
                    break
            if not match_re:
                raise ValueError(
                    f"OSLODatatypeUnion.objectUri is invalid. Value = '{union_datatype.objectUri}'")

        if union_datatype.name == '':
            raise ValueError(f"OSLODatatypeUnion.name is invalid. Value = '{union_datatype.name}'")

        return self.create_block_to_write_from_complex_primitive_or_union_types(oslo_datatype=union_datatype,
                                                                                type_field='UnionType',
                                                                                model_location=model_location)



import os
import sqlite3
from pathlib import Path
from sqlite3 import Connection

from otlmow_modelbuilder.SQLDataClasses.GeneralInfoRecord import GeneralInfoRecord
from otlmow_modelbuilder.SQLDataClasses.Inheritance import Inheritance
from otlmow_modelbuilder.SQLDataClasses.OSLOAttribuut import OSLOAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeComplex import OSLODatatypeComplex
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeComplexAttribuut import OSLODatatypeComplexAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypePrimitive import OSLODatatypePrimitive
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypePrimitiveAttribuut import OSLODatatypePrimitiveAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeUnion import OSLODatatypeUnion
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypeUnionAttribuut import OSLODatatypeUnionAttribuut
from otlmow_modelbuilder.SQLDataClasses.OSLOEnumeration import OSLOEnumeration
from otlmow_modelbuilder.SQLDataClasses.OSLORelatie import OSLORelatie
from otlmow_modelbuilder.SQLDataClasses.OSLOTypeLink import OSLOTypeLink


class OSLOInMemoryCreator:
    def __init__(self, path: Path):
        self.connection: Connection
        self.path: Path = path
        if path is None or path == '':
            self.file_exists = False
        else:
            self.path = path.resolve()
            self.file_exists = os.path.isfile(self.path)
            if not self.file_exists:
                raise FileNotFoundError(f"{self.path} is not a valid path. File does not exist.")

    def __enter__(self):
        self.connection = sqlite3.connect(self.path)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.connection.close()

    def perform_read_query(self, query: str, params=None) -> [tuple]:
        if params is None:
            params = {}

        cur = self.connection.cursor()
        return list(cur.execute(query, params))

    def get_general_info(self) -> [tuple]:
        data = self.perform_read_query("SELECT * FROM GeneralInfo")
        return [GeneralInfoRecord(row[0], row[1]) for row in data]

    def get_otl_version(self) -> str:
        data = self.perform_read_query("SELECT Waarde FROM GeneralInfo WHERE Parameter = 'Version'")

        return None if len(data) == 0 else data[0][0]

    def get_all_primitive_datatype_attributes(self) -> [OSLODatatypePrimitiveAttribuut]:
        data = self.perform_read_query(
            "SELECT name, label_nl, definition_nl, class_uri, kardinaliteit_min, kardinaliteit_max, uri, type, "
            "overerving, constraints, readonly, usagenote_nl, deprecated_version "
            "FROM OSLODatatypePrimitiveAttributen "
            "ORDER BY uri")

        return [OSLODatatypePrimitiveAttribuut(
            row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8], row[9], row[10], row[11], row[12])
            for row in data]

    def get_all_primitive_datatypes(self) -> [OSLODatatypePrimitive]:
        data = self.perform_read_query(
            "SELECT name, uri, definition_nl, label_nl, usagenote_nl, deprecated_version "
            "FROM OSLODatatypePrimitive "
            "ORDER BY uri")

        return [OSLODatatypePrimitive(row[0], row[1], row[2], row[3], row[4], row[5]) for row in data]

    def get_all_classes(self) -> [OSLOClass]:
        data = self.perform_read_query(
            "SELECT label_nl, name, uri, definition_nl, usagenote_nl, abstract, deprecated_version "
            "FROM OSLOClass "
            "ORDER BY uri")

        return [OSLOClass(row[0], row[1], row[2], row[3], row[4], row[5], row[6]) for row in data]

    def get_all_classes_and_class_dict(self) -> tuple[list[OSLOClass], dict[str, OSLOClass]]:
        data = self.perform_read_query(
            "SELECT label_nl, name, uri, definition_nl, usagenote_nl, abstract, deprecated_version "
            "FROM OSLOClass "
            "ORDER BY uri")

        class_dict = {}
        return [self._store_and_return_class_from_row(row, class_dict) for row in data], class_dict

    @staticmethod
    def _store_and_return_class_from_row(row: tuple, class_dict: dict) -> OSLOClass:
        class_ = OSLOClass(row[0], row[1], row[2], row[3], row[4], row[5], row[6])
        class_dict[class_.objectUri] = class_
        return class_

    def get_class_by_uri(self, class_uri: str) -> OSLOClass:
        data = self.perform_read_query(
            query="SELECT label_nl, name, uri, definition_nl, usagenote_nl, abstract, deprecated_version "
                  "FROM OSLOClass "
                  "WHERE uri=:uriclass",
            params={"uriclass": class_uri})

        class_results = [OSLOClass(row[0], row[1], row[2], row[3], row[4], row[5], row[6]) for row in data]

        if len(class_results) > 1:
            raise ValueError(f'There should only be 1 class with this uri: {class_uri}')

        if len(class_results) == 1:
            return class_results[0]

    def get_attributes_by_class_uri(self, class_uri, include_abstract=False) -> [OSLOAttribuut]:
        overerving_in_query = '' if include_abstract else 'overerving = 0 AND '
        data = self.perform_read_query(
            query="SELECT name, label_nl, definition_nl, class_uri, kardinaliteit_min, kardinaliteit_max, uri, type, "
                  "overerving, constraints, readonly, usagenote_nl, deprecated_version "
                  f"FROM OSLOAttributen WHERE class_uri=:uriclass AND {overerving_in_query}name <> 'typeURI' "
                  "ORDER BY uri",
            params={"uriclass": class_uri})

        return [OSLOAttribuut(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8], row[9], row[10],
                              row[11], row[12]) for row in data]

    def get_all_attributes(self, include_abstract=False) -> [OSLOAttribuut]:
        overerving_in_query = '' if include_abstract else 'overerving = 0 AND '
        data = self.perform_read_query(
            "SELECT name, label_nl, definition_nl, class_uri, kardinaliteit_min, kardinaliteit_max, uri, type, "
            "overerving, constraints, readonly, usagenote_nl, deprecated_version "
            f"FROM OSLOAttributen WHERE {overerving_in_query}name <> 'typeURI' "
            f"ORDER BY uri")

        return [OSLOAttribuut(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8], row[9], row[10],
                              row[11], row[12]) for row in data]

    def get_all_inheritances(self) -> [Inheritance]:
        data = self.perform_read_query(
            "SELECT base_name, base_uri, class_uri, class_name, deprecated_version "
            "FROM InternalBaseClass "
            "ORDER BY base_uri, class_uri")

        return [Inheritance(row[0], row[1], row[2], row[3], row[4]) for row in data]

    def get_all_complex_datatypes(self) -> [OSLODatatypeComplex]:
        data = self.perform_read_query(
            "SELECT name, uri, definition_nl, label_nl, usagenote_nl, deprecated_version "
            "FROM OSLODatatypeComplex "
            "ORDER BY uri")

        return [OSLODatatypeComplex(row[0], row[1], row[2], row[3], row[4], row[5]) for row in data]

    def get_all_complex_datatype_attributes(self) -> [OSLODatatypeComplexAttribuut]:
        data = self.perform_read_query(
            "SELECT name, label_nl, definition_nl, class_uri, kardinaliteit_min, kardinaliteit_max, uri, type, "
            "overerving, constraints, readonly, usagenote_nl, deprecated_version "
            "FROM OSLODatatypeComplexAttributen "
            "ORDER BY uri")

        return [OSLODatatypeComplexAttribuut(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8],
                                             row[9], row[10], row[11], row[12]) for row in data]

    def get_all_union_datatypes(self) -> [OSLODatatypeUnion]:
        data = self.perform_read_query(
            "SELECT name, uri, definition_nl, label_nl, usagenote_nl, deprecated_version "
            "FROM OSLODatatypeUnion "
            "ORDER BY uri")

        return [OSLODatatypeUnion(row[0], row[1], row[2], row[3], row[4], row[5]) for row in data]

    def get_all_union_datatype_attributes(self) -> [OSLODatatypeUnionAttribuut]:
        data = self.perform_read_query(
            "SELECT name, label_nl, definition_nl, class_uri, kardinaliteit_min, kardinaliteit_max, uri, type, "
            "overerving, constraints, readonly, usagenote_nl, deprecated_version "
            "FROM OSLODatatypeUnionAttributen "
            "ORDER BY uri")

        return [OSLODatatypeUnionAttribuut(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8],
                                           row[9], row[10], row[11], row[12]) for row in data]

    def get_all_enumerations(self) -> [OSLOEnumeration]:
        data = self.perform_read_query(
            "SELECT name, uri, usagenote_nl, definition_nl, label_nl, codelist, deprecated_version "
            "FROM OSLOEnumeration "
            "ORDER BY uri")

        return [OSLOEnumeration(row[0], row[1], row[2], row[3], row[4], row[5], row[6]) for row in data]

    def get_all_typelinks(self) -> [OSLOTypeLink]:
        data = self.perform_read_query(
            "SELECT item_uri, item_tabel, deprecated_version "
            "FROM TypeLinkTabel "
            "ORDER BY item_uri")

        return [OSLOTypeLink(row[0], row[1], row[2]) for row in data]

    def get_all_relations(self) -> [OSLORelatie]:
        data = self.perform_read_query(
            "SELECT * "
            "FROM OSLORelaties "
            "ORDER BY uri, bron_uri, doel_uri")

        return [OSLORelatie(row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7]) for row in data]

    def check_on_base_classes(self):
        data = self.perform_read_query(
            """
WITH valid_base_classes AS (
	VALUES ('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#RelatieObject'),
		('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Derdenobject'),
        ('http://purl.org/dc/terms/Agent'),
        ('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Toegangsprocedure'),
        ('https://wegenenverkeer.data.vlaanderen.be/ns/abstracten#AbstracteAanvullendeGeometrie'),
        ('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMObject')),    
concrete_classes AS (
	SELECT uri AS concrete_uri
	FROM OSLOClass o
	WHERE abstract = 0 AND concrete_uri NOT IN valid_base_classes),
inheritance_checks AS (
	SELECT concrete_uri, 
		 CASE WHEN inh7.base_uri IS NOT NULL AND inh7.base_uri IN valid_base_classes THEN 1
		 	WHEN inh6.base_uri IS NOT NULL AND inh6.base_uri IN valid_base_classes THEN 1
		 	WHEN inh5.base_uri IS NOT NULL AND inh5.base_uri IN valid_base_classes THEN 1
		 	WHEN inh4.base_uri IS NOT NULL AND inh4.base_uri IN valid_base_classes THEN 1
		 	WHEN inh3.base_uri IS NOT NULL AND inh3.base_uri IN valid_base_classes THEN 1
		 	WHEN inh2.base_uri IS NOT NULL AND inh2.base_uri IN valid_base_classes THEN 1
		 	WHEN inh1.base_uri IS NOT NULL AND inh1.base_uri IN valid_base_classes THEN 1
		 	ELSE 0 END has_valid_base
	FROM concrete_classes
		LEFT JOIN InternalBaseClass inh1 ON concrete_classes.concrete_uri = inh1.class_uri
		LEFT JOIN InternalBaseClass inh2 ON inh1.base_uri = inh2.class_uri
		LEFT JOIN InternalBaseClass inh3 ON inh2.base_uri = inh3.class_uri
		LEFT JOIN InternalBaseClass inh4 ON inh3.base_uri = inh4.class_uri
		LEFT JOIN InternalBaseClass inh5 ON inh4.base_uri = inh5.class_uri
		LEFT JOIN InternalBaseClass inh6 ON inh5.base_uri = inh6.class_uri
		LEFT JOIN InternalBaseClass inh7 ON inh6.base_uri = inh7.class_uri)
SELECT concrete_uri -- concrete_uri to view the list of classes that do not have a valid base class 
FROM inheritance_checks GROUP BY 1 HAVING sum(has_valid_base) = 0;""")
        return [] if len(data) == 0 else data[0][0]



from abc import ABC
from typing import Dict

from otlmow_modelbuilder.DatatypeBuilderFunctions import get_attributen_by_type_field
from otlmow_modelbuilder.GenericBuilderFunctions import add_attributen_to_data_block, \
    get_fields_to_import_from_list_of_attributes
from otlmow_modelbuilder.HelperFunctions import wrap_in_quotes, escape_backslash
from otlmow_modelbuilder.OSLOCollector import OSLOCollector


class AbstractDatatypeCreator(ABC):
    def __init__(self, oslo_collector: OSLOCollector):
        self.oslo_collector = oslo_collector

    def create_block_to_write_from_complex_primitive_or_union_types(self, oslo_datatype, type_field='',
                                                                    model_location='', valid_uri_and_types: Dict = None):
        if valid_uri_and_types is None:
            valid_uri_and_types = {}
        attributen = get_attributen_by_type_field(self.oslo_collector, type_field, oslo_datatype)

        datablock = ['# coding=utf-8',
                     'from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLAttribuut']

        list_fields_to_start_with = [f'{type_field}Field']
        if type_field == 'UnionType':
            list_fields_to_start_with.append('UnionWaarden')
        elif type_field == 'Complex':
            datablock.append('from otlmow_model.OtlmowModel.BaseClasses.WaardenObject import WaardenObject')
        elif type_field == 'Primitive' or type_field == 'KwantWrd':
            datablock.append('from otlmow_model.OtlmowModel.BaseClasses.OTLField import OTLField')
            datablock.append('from otlmow_model.OtlmowModel.BaseClasses.WaardenObject import WaardenObject')
            list_fields_to_start_with = []
        list_of_fields = get_fields_to_import_from_list_of_attributes(self.oslo_collector, attributen,
                                                                      valid_uri_and_types=valid_uri_and_types,
                                                                      list_to_start_from=list_fields_to_start_with)
        base_fields = ['BooleanField', 'ComplexField', 'DateField', 'DateTimeField', 'FloatOrDecimalField',
                       'IntegerField',
                       'KeuzelijstField', 'UnionTypeField', 'URIField', 'LiteralField', 'NonNegIntegerField',
                       'TimeField',
                       'StringField', 'UnionWaarden']
        for module in list_of_fields:
            if module not in base_fields:
                datablock.append(f'from ..Datatypes.{module} import {module}')
            else:
                datablock.append(f'from otlmow_model.OtlmowModel.BaseClasses.{module} import {module}')

        datablock.append('')
        datablock.append('')
        datablock.append(f'# Generated with {self.__class__.__name__}. To modify: extend, do not edit')
        if type_field == 'UnionType':
            datablock.append(f'class {oslo_datatype.name}Waarden(UnionWaarden):')
            datablock.append('    def __init__(self):')
            datablock.append('        UnionWaarden.__init__(self)')
        else:
            datablock.append(f'class {oslo_datatype.name}Waarden(WaardenObject):')
            datablock.append('    def __init__(self):')
            datablock.append('        WaardenObject.__init__(self)')

        add_attributen_to_data_block(attributen=attributen, datablock=datablock, type_field=type_field, valid_uri_and_types=valid_uri_and_types)

        if type_field == 'Primitive' or type_field == 'KwantWrd':
            type_field = 'OTL'

        datablock.append(''),
        datablock.append(f'# Generated with {self.__class__.__name__}. To modify: extend, do not edit')
        datablock.append(f'class {oslo_datatype.name}({type_field}Field):')
        datablock.append(f'    """{oslo_datatype.definition}"""')
        datablock.append(f'    naam = {wrap_in_quotes(oslo_datatype.name)}')
        datablock.append(f'    label = {wrap_in_quotes(oslo_datatype.label)}')
        datablock.append(f'    objectUri = {wrap_in_quotes(oslo_datatype.objectUri)}')
        datablock.append(f'    definition = {wrap_in_quotes(oslo_datatype.definition)}')
        if oslo_datatype.usagenote != '':
            if oslo_datatype.deprecated_version != '' and 'uit gebruik sinds versie' in oslo_datatype.usagenote:
                no_tag = oslo_datatype.deprecated_version.split('-')[0]
                oslo_datatype.usagenote = oslo_datatype.usagenote.replace(oslo_datatype.deprecated_version, no_tag)
            datablock.append(f'    usagenote = {escape_backslash(wrap_in_quotes(oslo_datatype.usagenote))}')
        if oslo_datatype.deprecated_version != '':
            oslo_datatype.deprecated_version = oslo_datatype.deprecated_version.split('-')[0]
            datablock.append(f'    deprecated_version = {wrap_in_quotes(oslo_datatype.deprecated_version)}'),
        if type_field == 'OTL':
            datablock.append('    waarde_shortcut_applicable = True')
        datablock.append(f'    waardeObject = {oslo_datatype.name}Waarden')
        datablock.append(f'')
        datablock.append(f'    def __str__(self):')
        datablock.append(f'        return {type_field}Field.__str__(self)')
        datablock.append('')

        return datablock




import logging

from otlmow_modelbuilder.AbstractDatatypeCreator import AbstractDatatypeCreator


class LegacyClassCreator(AbstractDatatypeCreator):
    def __init__(self):
        super().__init__(None)
        logging.info("Created an instance of LegacyClassCreator")


    def create_blocks_to_write_from_rows(self, lgc_type_row: [str], model_location='') -> [str]:
        try:
            legacy_uri = lgc_type_row[0]
            legacy_name = lgc_type_row[2]
            legacy_definition = lgc_type_row[1] + "\n\t" + lgc_type_row[3]
        except Exception as e:
            logging.error(f"Error in create_blocks_to_write_from_rows: {e}")
            raise e

        if '.' in legacy_name:
            legacy_name = legacy_name.replace('.', '_')
        if '-' in legacy_name:
            legacy_name = legacy_name.replace('-', '_')
        if legacy_name == "RIS":
            legacy_name = "RISLegacy"
        elif legacy_name == 'Fietstel':
            legacy_name = 'FietstelLegacy'
        elif legacy_name == 'Brug':
            legacy_name = 'BeweegbareBrug'
        elif legacy_name == 'Voedingskeuzeschakelaar':
            legacy_name = 'VKS'

        if legacy_uri == '':
            raise ValueError(f"OSLOClass.objectUri is invalid. Value = '{legacy_uri}'")

        if legacy_name == '':
            raise ValueError(f"OSLOClass.name is invalid. Value = '{legacy_name}'")

        return self.create_block_from_class(legacy_uri=legacy_uri, legacy_name=legacy_name,
                                            legacy_definition=legacy_definition, model_location=model_location)

    def create_block_from_class(self, legacy_uri: str, legacy_name: str, legacy_definition: str,
                                model_location: str = '') -> [str]:
        return [
            '# coding=utf-8',
            'from otlmow_model.OtlmowModel.BaseClasses.LegacyObject import LegacyObject',
            '',
            '',
            f'# Generated with {self.__class__.__name__}. To modify: extend, do not edit',
            f'class {legacy_name}(LegacyObject):',
            f'    """{legacy_definition}"""',
            '',
            f"    typeURI = '{legacy_uri}'",
            '    """De URI van het object volgens https://www.w3.org/2001/XMLSchema#anyURI."""',
            '',
            '    def __init__(self):',
            '        super().__init__()',
        ]



import concurrent.futures
import csv
import json
import logging
import os
import shutil
import subprocess
from collections import defaultdict
from concurrent.futures import FIRST_COMPLETED
from datetime import datetime
from os import path
from pathlib import Path
from typing import Dict

from tqdm import tqdm

from otlmow_modelbuilder.GenericBuilderFunctions import write_to_file
from otlmow_modelbuilder.GeometrieArtefactCollector import GeometrieArtefactCollector
from otlmow_modelbuilder.HelperFunctions import get_ns_and_name_from_uri, get_class_directory_from_ns, \
    get_titlecase_from_ns
from otlmow_modelbuilder.LegacyClassCreator import LegacyClassCreator
from otlmow_modelbuilder.OSLOCollector import OSLOCollector
from otlmow_modelbuilder.OTLClassCreator import OTLClassCreator
from otlmow_modelbuilder.OTLComplexDatatypeCreator import OTLComplexDatatypeCreator
from otlmow_modelbuilder.OTLEnumerationCreator import OTLEnumerationCreator
from otlmow_modelbuilder.OTLExtraChecker import OTLExtraChecker
from otlmow_modelbuilder.OTLPrimitiveDatatypeCreator import OTLPrimitiveDatatypeCreator
from otlmow_modelbuilder.OTLUnionDatatypeCreator import OTLUnionDatatypeCreator
from otlmow_modelbuilder.SQLDataClasses.Inheritance import Inheritance
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass


class OTLModelCreator:
    def __init__(self):
        pass

    @staticmethod
    def create_full_model(directory: Path, oslo_collector: OSLOCollector,
                          geo_artefact_collector: GeometrieArtefactCollector, settings: Dict, include_legacy: bool,
                          environment: str = '', include_kl_test_keuzelijst: bool = False):
        logging.info('started creating model at ' + datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
        OTLModelCreator.check_and_create_subdirectories(directory)
        oslo_collector.query_correct_base_classes()

        OTLModelCreator.check_for_attributes_with_different_case(oslo_collector)
        OTLModelCreator.check_for_nested_attributes_in_classes(collector=oslo_collector,
                                                               known_classes_uris=settings['nested_list_class_uris'])
        OTLModelCreator.create_primitive_datatypes(
            directory=directory / 'OtlmowModel', oslo_collector=oslo_collector,
            primitive_datatype_validation_rules=settings['primitive_datatype_validation_rules'])
        OTLModelCreator.create_complex_datatypes(
            directory=directory / 'OtlmowModel', oslo_collector=oslo_collector,
            complex_datatype_validation_rules=settings['complex_datatype_validation_rules'])
        OTLModelCreator.create_union_datatypes(
            directory=directory / 'OtlmowModel', oslo_collector=oslo_collector,
            union_datatype_validation_rules=settings['union_datatype_validation_rules'])
        OTLModelCreator.create_enumerations(
            directory=directory / 'OtlmowModel', environment=environment, oslo_collector=oslo_collector,
            enumeration_validation_rules=settings['enumeration_validation_rules'],
            include_kl_test_keuzelijst=include_kl_test_keuzelijst)
        OTLModelCreator.create_classes(
            directory=directory / 'OtlmowModel', oslo_collector=oslo_collector,
            geo_artefact_collector=geo_artefact_collector,
            valid_uri_and_types=settings['complex_datatype_validation_rules']['valid_uri_and_types'])
        if include_legacy:
            oslo_collector = OTLModelCreator.create_legacy_classes(oslo_collector=oslo_collector,
                directory=directory / 'OtlmowModel', legacy_types_path=Path(__file__).parent / 'legacy_types.csv')
        OTLExtraChecker.modify_for_extra_checks(directory=directory / 'OtlmowModel')
        OTLModelCreator.add_generated_info(
            directory=directory / 'OtlmowModel', oslo_collector=oslo_collector)
        logging.info(f'finished creating model in {directory} at ' + datetime.now().strftime("%d/%m/%Y %H:%M:%S"))

    @staticmethod
    def create_primitive_datatypes(directory, oslo_collector, primitive_datatype_validation_rules):
        creator = OTLPrimitiveDatatypeCreator(oslo_collector)

        for prim_datatype in tqdm(oslo_collector.primitive_datatypes):
            if prim_datatype.objectUri in ['http://www.w3.org/2000/01/rdf-schema#Literal',
                                           'http://www.w3.org/2001/XMLSchema#dateTime',
                                           'http://www.w3.org/2001/XMLSchema#integer',
                                           'http://www.w3.org/2001/XMLSchema#decimal',
                                           'http://www.w3.org/2001/XMLSchema#time',
                                           'http://www.w3.org/2001/XMLSchema#date',
                                           'http://www.w3.org/2001/XMLSchema#nonNegativeInteger',
                                           'http://www.w3.org/2001/XMLSchema#string',
                                           'http://www.w3.org/2001/XMLSchema#boolean',
                                           'http://www.w3.org/2001/XMLSchema#anyURI']:
                logging.info(f"Skip creating class for {prim_datatype.name}")
                continue

            try:
                model_name = OTLModelCreator.get_model_name_from_directory_path(directory)
                data_to_write = creator.create_block_to_write_from_primitive_types(
                    prim_datatype, model_location=model_name,
                    primitive_datatype_validation_rules=primitive_datatype_validation_rules)
                if data_to_write is None:
                    logging.info(f"Could not create a class for {prim_datatype.name}")
                if len(data_to_write) == 0:
                    logging.info(f"Could not create a class for {prim_datatype.name}")
                write_to_file(prim_datatype.name, 'Datatypes', data_to_write, relative_path=directory)
            except BaseException as e:
                logging.error(str(e))
                logging.error(f"Could not create a class for {prim_datatype.name}")

    @staticmethod
    def get_model_name_from_directory_path(directory):
        if directory.name == 'TestClasses':
            return f'{directory.parent.name}.{directory.name}'
        else:
            return directory.name

    @staticmethod
    def create_complex_datatypes(directory, oslo_collector, complex_datatype_validation_rules):
        creator = OTLComplexDatatypeCreator(oslo_collector)

        for complex_datatype in tqdm(oslo_collector.complex_datatypes):
            try:
                model_name = OTLModelCreator.get_model_name_from_directory_path(directory)
                data_to_write = creator.create_block_to_write_from_complex_types(
                    complex_datatype, complex_datatype_validation_rules=complex_datatype_validation_rules,
                    model_location=model_name)
                if data_to_write is None:
                    logging.info(f"Could not create a class for {complex_datatype.name}")
                if len(data_to_write) == 0:
                    logging.info(f"Could not create a class for {complex_datatype.name}")
                write_to_file(complex_datatype.name, 'Datatypes', data_to_write, relative_path=directory)
            except BaseException as e:
                logging.error(str(e))
                logging.error(f"Could not create a class for {complex_datatype.name}")

    @staticmethod
    def create_union_datatypes(directory, oslo_collector, union_datatype_validation_rules):
        creator = OTLUnionDatatypeCreator(oslo_collector)

        for union_datatype in tqdm(oslo_collector.union_datatypes):
            try:
                model_name = OTLModelCreator.get_model_name_from_directory_path(directory)
                data_to_write = creator.create_block_to_write_from_union_types(
                    union_datatype, model_location=model_name,
                    union_datatype_validation_rules=union_datatype_validation_rules)
                if data_to_write is None:
                    logging.info(f"Could not create a class for {union_datatype.name}")
                if len(data_to_write) == 0:
                    logging.info(f"Could not create a class for {union_datatype.name}")
                write_to_file(union_datatype.name, 'Datatypes', data_to_write, relative_path=directory)
            except BaseException as e:
                logging.error(str(e))
                logging.error(f"Could not create a class for {union_datatype.name}")

    @staticmethod
    def create_enumerations(directory, oslo_collector, enumeration_validation_rules, environment: str = '',
                            include_kl_test_keuzelijst: bool = False):
        with OTLEnumerationCreator(oslo_collector, env=environment,
                                        include_kl_test_keuzelijst=include_kl_test_keuzelijst) as creator:

            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = {executor.submit(OTLModelCreator.create_enumeration, creator=creator, directory=directory,
                                           enumeration=enumeration, environment=environment,
                                           enumeration_validation_rules=enumeration_validation_rules): enumeration for
                           enumeration in oslo_collector.enumerations}
                with tqdm(total=len(futures)) as pbar:
                    while futures:
                        new_futures = {}
                        done, pending = concurrent.futures.wait(futures, return_when=FIRST_COMPLETED, timeout=60)
                        for fut in done:
                            if fut.exception():
                                enumeration = futures[fut]
                                new_futures[executor.submit(
                                    OTLModelCreator.create_enumeration, creator=creator, directory=directory,
                                    enumeration=enumeration, environment=environment,
                                    enumeration_validation_rules=enumeration_validation_rules)
                                ] = job
                            else:
                                pbar.update()
                        for fut in pending:
                            job = futures[fut]
                            new_futures[fut] = job
                        futures = new_futures

    @staticmethod
    def create_enumeration(creator, directory, enumeration, environment, enumeration_validation_rules):
        try:
            data_to_write = creator.create_block_to_write_from_enumerations(
                enumeration, environment=environment, enumeration_validation_rules=enumeration_validation_rules)
            if data_to_write is None:
                logging.error(f"Could not create a class for {enumeration.name}")
            if len(data_to_write) == 0:
                logging.error(f"Could not create a class for {enumeration.name}")
            write_to_file(enumeration.name, 'Datatypes', data_to_write, relative_path=directory)
        except BaseException as e:
            logging.error(str(e))
            logging.error(f"Could not create a class for {enumeration.name}")

    @staticmethod
    def create_classes(directory, oslo_collector, geo_artefact_collector, valid_uri_and_types):
        creator = OTLClassCreator(oslo_collector, geo_artefact_collector)

        for oslo_class in tqdm(oslo_collector.classes):
            try:
                model_name = OTLModelCreator.get_model_name_from_directory_path(directory)
                data_to_write = creator.create_blocks_to_write_from_classes(oslo_class, model_location=model_name,
                                                                            valid_uri_and_types=valid_uri_and_types)
                if data_to_write is None:
                    logging.info(f"Could not create a class for {oslo_class.name}")
                if len(data_to_write) == 0:
                    logging.info(f"Could not create a class for {oslo_class.name}")
                class_directory = 'Classes'
                ns = None
                if oslo_class.objectUri != 'http://purl.org/dc/terms/Agent':
                    ns, _ = get_ns_and_name_from_uri(oslo_class.objectUri)
                if ns is not None:
                    class_directory = get_class_directory_from_ns(ns)

                write_to_file(oslo_class.name, class_directory, data_to_write, relative_path=directory)
            except Exception as e:
                logging.error(str(e))
                logging.error(f"Could not create a class for {oslo_class.name}")

        data_to_write = []
        for oslo_class in oslo_collector.classes:
            if oslo_class.abstract:
                continue
            ns = None
            if oslo_class.objectUri != 'http://purl.org/dc/terms/Agent':
                ns, name = get_ns_and_name_from_uri(oslo_class.objectUri)
            else:
                name = 'Agent'

            if ns is not None:
                data_to_write.append(f'from ..Classes.{get_titlecase_from_ns(ns)}.{name} import {name}')
            else:
                data_to_write.append(f'from ..Classes.{name} import {name}')

        write_to_file('all_classes', 'Helpers', data_to_write, relative_path=directory)


    @staticmethod
    def check_and_create_subdirectories(directory: Path):
        if not path.exists(directory):
            os.makedirs(directory)
        if not path.exists(directory):
            raise OSError(f'The directory {directory} does not exist. Please create it first.')
        if not path.isdir(directory):
            raise NotADirectoryError(f'{directory} is not a directory.')

        if not path.exists(directory / 'OtlmowModel'):
            os.mkdir(directory / 'OtlmowModel')

        model_directory = directory / 'OtlmowModel'

        OTLModelCreator.clean_directory(model_directory)
        OTLModelCreator.copy_fixed_classes_from_otlmow_model(model_directory)

        if not path.exists(model_directory / 'Classes'):
            os.mkdir(model_directory / 'Classes')
        if not path.exists(model_directory / 'Datatypes'):
            os.mkdir(model_directory / 'Datatypes')

        if not path.exists(model_directory / 'Classes/Abstracten'):
            os.mkdir(model_directory / 'Classes/Abstracten')
        if not path.exists(model_directory / 'Classes/ImplementatieElement'):
            os.mkdir(model_directory / 'Classes/ImplementatieElement')
        if not path.exists(model_directory / 'Classes/Installatie'):
            os.mkdir(model_directory / 'Classes/Installatie')
        if not path.exists(model_directory / 'Classes/Legacy'):
            os.mkdir(model_directory / 'Classes/Legacy')
        if not path.exists(model_directory / 'Classes/Levenscyclus'):
            os.mkdir(model_directory / 'Classes/Levenscyclus')
        if not path.exists(model_directory / 'Classes/Onderdeel'):
            os.mkdir(model_directory / 'Classes/Onderdeel')
        if not path.exists(model_directory / 'Classes/ProefEnMeting'):
            os.mkdir(model_directory / 'Classes/ProefEnMeting')

    @staticmethod
    def clean_directory(directory):
        for root, dirs, files in os.walk(directory):
            for direc in dirs:
                shutil.rmtree(os.path.join(root, direc))
            for file in files:
                os.remove(os.path.join(root, file))

    @staticmethod
    def check_for_nested_attributes_in_classes(collector: OSLOCollector, exceptions=None,
                                               known_classes_uris: [str] = None):
        if known_classes_uris is None:
            known_classes_uris = []
        if exceptions is None:
            exceptions = [
                'https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#AllCasesTestClass.testComplexTypeMetKard']
        for cl in collector.classes:
            if cl.abstract:
                continue

            attributes = collector.find_attributes_by_class(cl)
            for attr in attributes:
                if attr.objectUri in exceptions:
                    # exception for test classes
                    continue
                list_found = False
                if attr.kardinaliteit_max != '1':
                    list_found = True
                nested_attrs = collector.find_complex_datatype_attributes_by_class_uri(attr.type)
                if len(nested_attrs) == 0:
                    nested_attrs = collector.find_primitive_datatype_attributes_by_class_uri(attr.type)
                if len(nested_attrs) == 0:
                    nested_attrs = collector.find_union_datatype_attributes_by_class_uri(attr.type)

                if len(nested_attrs) > 1:
                    try:
                        OTLModelCreator.check_for_nested_attributes_in_attributes(list_found, nested_attrs, collector)
                    except NotImplementedError as exc:
                        if cl.objectUri not in known_classes_uris:
                            raise NotImplementedError(f'found in {cl.objectUri} {attr.objectUri}') from exc

    @staticmethod
    def check_for_nested_attributes_in_attributes(list_already_found, attributes, collector):
        for attr in attributes:
            list_found = False
            if attr.kardinaliteit_max != '1':
                list_found = True

            if list_found and list_already_found:
                raise NotImplementedError(f'Found nested list attributes: {attr.objectUri}')
            nested_attrs = collector.find_complex_datatype_attributes_by_class_uri(attr.type)
            if len(nested_attrs) == 0:
                nested_attrs = collector.find_primitive_datatype_attributes_by_class_uri(attr.type)
            if len(nested_attrs) == 0:
                nested_attrs = collector.find_union_datatype_attributes_by_class_uri(attr.type)

            if len(nested_attrs) > 1:
                list_found = list_found or list_already_found
                OTLModelCreator.check_for_nested_attributes_in_attributes(list_found, nested_attrs, collector)

    @classmethod
    def check_for_attributes_with_different_case(cls, oslo_collector):
        lower_case_names = [d.name.lower() for d in oslo_collector.attributes]
        # find names that are identical when case-insensitive
        unique_lower_case_names = set(lower_case_names)
        problems = defaultdict(set)
        for lower_case_name in unique_lower_case_names:
            attrs = [d for d in oslo_collector.attributes if
                     d.name.lower() == lower_case_name and d.deprecated_version == '']
            if len(attrs) < 2:
                continue
            first_attr = attrs[0].name
            for attr in attrs[1:]:
                if attr.name != first_attr:
                    problems[first_attr.lower()].add(attrs[0].objectUri)
                    problems[first_attr.lower()].add(attr.objectUri)

        problems = {k: list(v) for k, v in problems.items()}

        # remove attributes that are known to be different
        known_list = ['basisoppervlakte', 'ipadres', 'risicoanalyse', 'technischefiche', 'opstelhoogte',
                      'buitendiameter', 'dnsnaam', 'funderingsaanzetonderdebodemvandewaterweg', 'binnendiameter',
                      'netwerktype', 'beschoeiingslengte', 'softwareversie', 'folietype', 'lamptype']
        for known in known_list:
            if known in problems:
                del problems[known]

        if problems:
            raise NotImplementedError(f'Found attributes with different case:\n{json.dumps(problems, indent=4)}')

    @classmethod
    def copy_fixed_classes_from_otlmow_model(cls, model_directory: Path):
        otlmow_model_github_dir = Path(__file__).parent / 'OTLMOW-Model'
        subprocess.call(["git", "clone", "--depth=1", "https://github.com/davidvlaminck/OTLMOW-Model/",
                         str(otlmow_model_github_dir.absolute())])

        otlmow_model_dir = otlmow_model_github_dir / 'otlmow_model' / 'OtlmowModel'
        if not otlmow_model_dir.exists():
            raise ModuleNotFoundError("Could not find otlmow-model directory")

        shutil.copytree(otlmow_model_dir / 'BaseClasses', model_directory / 'BaseClasses')
        shutil.copytree(otlmow_model_dir / 'Exceptions', model_directory / 'Exceptions')
        shutil.copytree(otlmow_model_dir / 'GeometrieTypes', model_directory / 'GeometrieTypes')
        shutil.copytree(otlmow_model_dir / 'Helpers', model_directory / 'Helpers')
        shutil.copytree(otlmow_model_dir / 'warnings', model_directory / 'warnings')

        try:
            shutil.rmtree(otlmow_model_github_dir)
        except PermissionError:
            print(f"Could not remove {otlmow_model_github_dir}")

    @classmethod
    def add_generated_info(cls, directory: Path, oslo_collector: OSLOCollector) -> None:
        relation_dict = cls.generate_relation_dict(oslo_collector)
        class_dict = cls.generate_class_dict(oslo_collector)
        with open(directory / 'generated_info.json', mode='w') as generated_info_file:
            json.dump({'relations': relation_dict, 'classes': class_dict}, generated_info_file, indent=4)

    @classmethod
    def generate_relation_dict(cls, oslo_collector: OSLOCollector) -> dict:
        relation_dict = {}
        for inheritance in oslo_collector.inheritances:
            if inheritance.base_uri == ('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#'
                                        'DirectioneleRelatie'):
                relation_dict[inheritance.class_uri] = {'directional': True}
            elif inheritance.base_uri == ('https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#'
                                          'NietDirectioneleRelatie'):
                relation_dict[inheritance.class_uri] = {'directional': False}
        return relation_dict

    @classmethod
    def generate_class_dict(cls, oslo_collector: OSLOCollector) -> dict:
        return {uri: {
                'abstract': oslo_class.abstract == 1,
                'name': oslo_class.name,
                'label': oslo_class.label,
                'deprecated_version': oslo_class.deprecated_version,
                'direct_subclasses': list(oslo_collector.find_subclasses_uri_by_class_uri(oslo_class.objectUri))
            } for uri, oslo_class in oslo_collector.class_dict.items()}

    @classmethod
    def create_legacy_classes(cls, oslo_collector: OSLOCollector, directory: Path, legacy_types_path: Path):
        creator = LegacyClassCreator()

        lgc_type_data = []
        with open(legacy_types_path) as legacy_types_file:
            lgc_type_data.extend(iter(csv.reader(legacy_types_file, delimiter='\t')))

        oslo_collector.inheritances.append(Inheritance(
            base_name='AIMDBStatus',
            base_uri='https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMDBStatus',
            class_name='LegacyObject',
            class_uri='https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject'
        ))
        oslo_collector.inheritances.append(Inheritance(
            base_name='AIMToestand',
            base_uri='https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMToestand',
            class_name='LegacyObject',
            class_uri='https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject'
        ))
        oslo_collector.inheritances.append(Inheritance(
            base_name='AIMVersie',
            base_uri='https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMVersie',
            class_name='LegacyObject',
            class_uri='https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject'
        ))

        lgc_class = OSLOClass(
            label='Legacy Object',
            name='LegacyObject',
            objectUri='https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject',
            deprecated_version='',
            definition='Abstracte voor een Legacy object',
            abstract=1
        )
        oslo_collector.classes.append(lgc_class)
        oslo_collector.class_dict['https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject'] = lgc_class


        for lgc_type_row in tqdm(lgc_type_data):
            class_name = lgc_type_row[2]
            class_uri = lgc_type_row[0]
            if '.' in class_name:
                class_name = class_name.replace('.', '_')
            if '-' in class_name:
                class_name = class_name.replace('-', '_')
            if class_name == "RIS":
                class_name = "RISLegacy"
            elif class_name == 'Fietstel':
                class_name = 'FietstelLegacy'
            elif class_name == 'Brug':
                class_name = 'BeweegbareBrug'
            elif class_name == 'Voedingskeuzeschakelaar':
                class_name = 'VKS'

            oslo_class = OSLOClass(
                objectUri=class_uri,
                name=class_name,
                label=lgc_type_row[1],
                deprecated_version='',
                abstract=False)
            oslo_collector.classes.append(oslo_class)
            oslo_collector.inheritances.append(Inheritance(
                base_name='LegacyObject',
                base_uri='https://lgc.data.wegenenverkeer.be/ns/installatie#LegacyObject',
                class_name=class_name,
                class_uri=class_uri
            ))
            oslo_collector.class_dict[class_uri] = oslo_class
            try:
                model_name = OTLModelCreator.get_model_name_from_directory_path(directory)
                data_to_write = creator.create_blocks_to_write_from_rows(lgc_type_row=lgc_type_row, model_location=model_name)
                if data_to_write is None:
                    logging.info(f"Could not create a class for {class_name}")
                if len(data_to_write) == 0:
                    logging.info(f"Could not create a class for {class_name}")
                class_directory = 'Classes/Legacy'

                write_to_file(class_name, class_directory, data_to_write, relative_path=directory)
            except Exception as e:
                logging.error(str(e))
                logging.error(f"Could not create a class for {class_name}")

        data_to_write = []
        for lgc_type_row in lgc_type_data:
            uri = lgc_type_row[0]
            ns, name = get_ns_and_name_from_uri(uri)
            data_to_write.append(f'from ..Classes.{get_titlecase_from_ns(ns)}.{name} import {name}')

        write_to_file('all_classes', 'Helpers', data_to_write, relative_path=directory, write_mode='a')

        return oslo_collector




import dataclasses


@dataclasses.dataclass
class OSLODatatypeComplex:
    name: str = ''
    objectUri: str = ''
    definition: str = ''
    label: str = ''
    usagenote: str = ''
    deprecated_version: str = ''







import os
from typing import List, Dict

from otlmow_modelbuilder.DatatypeBuilderFunctions import get_single_field_from_type_uri, get_type_link_from_attribuut, \
    get_type_name_of_complex_attribuut, get_type_name_of_union_attribuut, get_field_name_from_type_uri, \
    get_non_single_field_from_type_uri
from otlmow_modelbuilder.HelperFunctions import wrap_in_quotes, escape_backslash


def get_white_space_equivalent(string):
    return ''.join(' ' * len(string))


def import_datetime(datablock, var_datetime):
    for index, line in enumerate(datablock):
        if line == '':
            datablock.insert(1, f'from datetime import {var_datetime}')
            break
        if 'datetime' in line:
            datablock[index] = line + ', ' + var_datetime
            break


def cardinality_check(attribute, type_string, datablock):
    if attribute.kardinaliteit_max != '1':
        for index, line in enumerate(datablock):
            if line == '':
                datablock.insert(1, 'from typing import List')
                break
            if line == 'from typing import List':
                break
        return f'List[{type_string}]'
    else:
        return type_string


def get_type_hint_from_field(attribute, datablock, valid_uri_and_types):
    field_name = get_single_field_from_type_uri(attribute.type)
    if field_name == 'StringField' or field_name == 'URIField':
        return f' -> {cardinality_check(attribute, "str", datablock)}'
    elif field_name == 'FloatOrDecimalField':
        return f' -> {cardinality_check(attribute, "float", datablock)}'
    elif field_name == 'BooleanField':
        return f' -> {cardinality_check(attribute, "bool", datablock)}'
    elif field_name in ['IntegerField', 'NonNegIntegerField']:
        return f' -> {cardinality_check(attribute, "int", datablock)}'
    elif field_name == 'DateField':
        import_datetime(datablock, 'date')
        return f' -> {cardinality_check(attribute, "date", datablock)}'
    elif field_name == 'DateTimeField':
        import_datetime(datablock, 'datetime')
        return f' -> {cardinality_check(attribute, "datetime", datablock)}'
    elif field_name == 'TimeField':
        import_datetime(datablock, 'time')
        return f' -> {cardinality_check(attribute, "time", datablock)}'
    else:
        field_name = get_non_single_field_from_type_uri(attribute.type, valid_uri_and_types)
        if field_name[0] == 'KeuzelijstField':
            return f' -> {cardinality_check(attribute, "str", datablock)}'
        if field_name[0] == 'ComplexField' or field_name[0] == 'UnionTypeField':
            for index, line in enumerate(datablock):
                if line[0] not in ['#', 'f']:  # only loop over the import part
                    break
                if field_name[1] not in line:
                    continue
                if f'{field_name[1]}Waarden' in line:
                    break
                datablock[index] = line + f', {field_name[1]}Waarden'
                break
            return f' -> {cardinality_check(attribute, field_name[1] + "Waarden", datablock)}'
        else:
            print('not complexe waarde')
    return ''


def add_attributen_to_data_block(attributen, datablock: List[str], valid_uri_and_types: Dict, for_class_use=False,
                                 type_field=''):
    prop_datablock = []
    for attribuut in sorted(attributen, key=lambda a: a.name):
        if attribuut.overerving == 1:
            raise NotImplementedError(f"overerving 1 is not implemented, found in {attributen.objectUri}")

        whitespace = get_white_space_equivalent(f'        self._{attribuut.name} = OTLAttribuut(')
        field_name = get_single_field_from_type_uri(attribuut.type)

        if attribuut.objectUri == 'https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#Segmentcontroller.beveil?igingssleutel':
            attribuut.objectUri = attribuut.objectUri.replace('?', '_')
            attribuut.name = attribuut.name.replace('?', '_')

        datablock.append(f'        self._{attribuut.name} = OTLAttribuut(field={field_name},')
        datablock.append(f'{whitespace}naam={wrap_in_quotes(attribuut.name)},')
        datablock.append(f'{whitespace}label={wrap_in_quotes(attribuut.label)},')
        datablock.append(f'{whitespace}objectUri={wrap_in_quotes(attribuut.objectUri)},')

        if attribuut.usagenote != '':
            if attribuut.deprecated_version != '' and 'uit gebruik sinds versie' in attribuut.usagenote:
                no_tag = attribuut.deprecated_version.split('-')[0]
                attribuut.usagenote = attribuut.usagenote.replace(attribuut.deprecated_version, no_tag)
            datablock.append(f'{whitespace}usagenote={escape_backslash(wrap_in_quotes(attribuut.usagenote))},')
        if attribuut.readonly == 1:
            datablock.append(f'{whitespace}readonly=True,')
        if attribuut.deprecated_version != '':
            attribuut.deprecated_version = attribuut.deprecated_version.split('-')[0]
            datablock.append(f'{whitespace}deprecated_version={wrap_in_quotes(attribuut.deprecated_version)},')
        if attribuut.constraints != '':
            datablock.append(f'{whitespace}constraints={wrap_in_quotes(attribuut.constraints)},')
        if attribuut.kardinaliteit_min != '1':
            datablock.append(f'{whitespace}kardinaliteit_min={wrap_in_quotes(attribuut.kardinaliteit_min)},')
        if attribuut.kardinaliteit_max != '1':
            datablock.append(f'{whitespace}kardinaliteit_max={wrap_in_quotes(attribuut.kardinaliteit_max)},')
        definitie = wrap_in_quotes(attribuut.definition.replace('\n', ''))
        datablock.append(f'{whitespace}definition={definitie},')
        datablock.append(f'{whitespace}owner=self)')
        datablock.append('')

        owner_self = ', owner=self'
        if not for_class_use:
            owner_self += '._parent'

        type_hint = get_type_hint_from_field(attribuut, datablock, valid_uri_and_types)

        prop_datablock.append(f'    @property'),
        prop_datablock.append(f'    def {attribuut.name}(self){type_hint}:'),
        prop_datablock.append(f'        """{attribuut.definition}"""'),
        if type_field == 'KwantWrd' and attribuut.name == 'standaardEenheid':
            prop_datablock.append(f'        return self._{attribuut.name}.usagenote.split(\'"\')[1]'),
        else:
            prop_datablock.append(f'        return self._{attribuut.name}.get_waarde()'),
        prop_datablock.append(f''),
        if not attribuut.readonly:
            prop_datablock.append(f'    @{attribuut.name}.setter'),
            prop_datablock.append(f'    def {attribuut.name}(self, value):'),

            prop_datablock.append(f'        self._{attribuut.name}.set_waarde(value{owner_self})'),
            if type_field == 'UnionType':
                prop_datablock.append(f'        if value is not None:')
                prop_datablock.append(f"            self.clear_other_props('_{attribuut.name}')")
            prop_datablock.append(f'')

    datablock.extend(prop_datablock)

    return datablock


def get_fields_to_import_from_list_of_attributes(oslo_collector, attributen, valid_uri_and_types,
                                                 list_to_start_from=None):
    if list_to_start_from is None:
        list_to_start_from = []
    if len(attributen) == 0:
        return list_to_start_from

    collected_list = []
    collected_list.extend(list_to_start_from)

    for attribuut in attributen:
        type_link = get_type_link_from_attribuut(oslo_collector, attribuut).item_tabel
        if type_link == "OSLOEnumeration":
            collected_list.append(attribuut.type.split('#')[1])
        elif type_link == "OSLODatatypePrimitive":
            collected_list.append(get_single_field_from_type_uri(attribuut.type))
        elif type_link == "OSLODatatypeComplex":
            collected_list.append(get_type_name_of_complex_attribuut(attribuut.type,
                                                                     valid_uri_and_types=valid_uri_and_types))
        elif type_link == "OSLODatatypeUnion":
            collected_list.append(get_type_name_of_union_attribuut(attribuut.type))
        else:
            raise not NotImplementedError(f"{type_link} not implemented")

    distinct_types_list = list(set(collected_list))
    sorted_list = sorted(distinct_types_list, key=lambda t: t)
    return sorted_list


def write_to_file(filename, directory: str, data_to_write: List[str], relative_path='', write_mode='w'):
    if relative_path == '':
        base_dir = os.path.dirname(os.path.realpath(__file__))
        base_dir = os.path.abspath(os.path.join(base_dir, os.pardir))
    else:
        base_dir = relative_path
    path = f"{base_dir}/{directory}/{filename}.py"

    with open(path, write_mode, encoding='utf-8') as file:
        for line in data_to_write:
            file.write(line + "\n")



def get_ns_and_name_from_uri(objectUri):
    if '/ns/' not in objectUri:
        raise ValueError(f"{objectUri} is not a valid uri because it does not define a namespace ('/ns' is missing)")

    if '#' not in objectUri:
        raise ValueError(f"{objectUri} is not a valid uri because can not be split into a ns and name ('#' is missing)")

    short_uri = objectUri.split('/ns/')[1]
    short_uri_array = short_uri.split('#')
    return short_uri_array[0], short_uri_array[1]


def get_class_directory_from_ns(ns):
    return 'Classes/' + get_titlecase_from_ns(ns)


def get_titlecase_from_ns(ns: str) -> str:
    if ns == 'abstracten':
        return 'Abstracten'
    elif ns == 'implementatieelement':
        return 'ImplementatieElement'
    elif ns == 'installatie':
        return 'Installatie'
    elif ns == 'levenscyclus':
        return 'Levenscyclus'
    elif ns == 'onderdeel':
        return 'Onderdeel'
    elif ns == 'proefenmeting':
        return 'ProefEnMeting'
    else:
        raise ValueError()


def escape_backslash(text: str) -> str:
    return text.replace('\\', '\\\\')


def wrap_in_quotes(text: str) -> str:
    if not isinstance(text, str):
        raise TypeError
    if not text:
        return "''"
    singles = text.count("'")
    doubles = text.count('"')
    if singles > doubles:
        return '"' + text.replace('"', '\\"') + '"' if doubles > 0 else f'"{text}"'
    return "'" + text.replace("'", "\\'") + "'" if singles > 0 else f"'{text}'"



import copy

from otlmow_modelbuilder.GeometrieType import GeometrieType
from otlmow_modelbuilder.SQLDataClasses.Inheritance import Inheritance
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass


class GeometrieInheritanceProcessor:
    """This class uses the geometry type definitions and OTL classes and inheritances to determine what classes can inherit their geometry type definition instead of directly. The result is a changed set of geometry type definitions that contains both concrete and base classes. Inheritance works from at least one geometry type definition."""
    def __init__(self, geometrie_types: [GeometrieType], inheritances: [Inheritance], classes: [OSLOClass]):
        self.geometrie_types = geometrie_types
        self.inheritances = copy.copy(inheritances)
        self.classes = copy.copy(classes)

    def process_inheritances(self) -> [GeometrieType]:
        while len(self.inheritances) > 0:
            # zoek een inheritance en gebruik die voor een goede base class af te leiden

            inheritance = next(c for c in self.inheritances)
            base = self.search_for_valid_base_using_inheritance(inheritance.base_uri)
            self.process_inheritance_for_base(base)
        return self.geometrie_types

    def search_for_valid_base_using_inheritance(self, base):
        inheritances = list(filter(lambda i: i.base_uri == base, self.inheritances))
        if len(inheritances) == 0:
            return base
        for inheritance in inheritances:
            sub_inheritances = list(filter(lambda i: i.base_uri == inheritance.class_uri, self.inheritances))
            if len(sub_inheritances) > 0:
                return self.search_for_valid_base_using_inheritance(inheritance.class_uri)
        return base

    def process_inheritance_for_base(self, base):
        inheritances = list(filter(lambda i: i.base_uri == base, self.inheritances))
        if len(inheritances) == 0:
            return

        no_None_types = True
        geen = 1
        point = 1
        line = 1
        polygon = 1

        # for all inheritances, find the common geometry types
        # if any class does not have a geometry tpe (likely due to lack of inheritance), don't inherit anything
        for inheritance in inheritances:
            geo_type = next((g for g in self.geometrie_types if g.objectUri == inheritance.class_uri), None)
            if geo_type is None:
                no_None_types = False
                break
            if geen == 1 and geo_type.geen_geometrie == 0:
                geen = 0
            if point == 1 and geo_type.punt3D == 0:
                point = 0
            if line == 1 and geo_type.lijn3D == 0:
                line = 0
            if polygon == 1 and geo_type.polygoon3D == 0:
                polygon = 0

        # if there is something to inherit: create and add a new geometry Type (using base) and remove others
        if no_None_types and (geen + point + line + polygon > 0):
            new_geo_type = GeometrieType()
            new_geo_type.objectUri = base
            cls = next((g for g in self.classes if g.objectUri == base), None)
            if cls is not None:
                new_geo_type.label_nl = cls.label
            new_geo_type.geen_geometrie = geen
            new_geo_type.punt3D = point
            new_geo_type.lijn3D = line
            new_geo_type.polygoon3D = polygon

            self.geometrie_types.append(new_geo_type)
            for inheritance in inheritances:
                current_geo_type = next((g for g in self.geometrie_types if g.objectUri == inheritance.class_uri), None)
                if current_geo_type is not None:
                    self.geometrie_types.remove(current_geo_type)

        # regardless of outcomes, remove all inheritances from base to avoid duplicates
        for inheritance in inheritances:
            self.inheritances.remove(inheritance)




import logging
import re
from typing import Dict

from otlmow_modelbuilder.GenericBuilderFunctions import add_attributen_to_data_block, \
    get_fields_to_import_from_list_of_attributes
from otlmow_modelbuilder.GeometrieArtefactCollector import GeometrieArtefactCollector
from otlmow_modelbuilder.GeometrieInheritanceProcessor import GeometrieInheritanceProcessor
from otlmow_modelbuilder.GeometrieType import GeometrieType
from otlmow_modelbuilder.HelperFunctions import get_ns_and_name_from_uri, get_class_directory_from_ns
from otlmow_modelbuilder.AbstractDatatypeCreator import AbstractDatatypeCreator
from otlmow_modelbuilder.SQLDataClasses.Inheritance import Inheritance
from otlmow_modelbuilder.SQLDataClasses.OSLOClass import OSLOClass
from otlmow_modelbuilder.OSLOCollector import OSLOCollector


class OTLClassCreator(AbstractDatatypeCreator):
    def __init__(self, oslo_collector: OSLOCollector, geo_artefact_collector: GeometrieArtefactCollector = None):
        super().__init__(oslo_collector)
        logging.info("Created an instance of OTLClassCreator")
        self.oslo_collector = oslo_collector
        self.geoACollector = geo_artefact_collector
        self.geometry_types = []

        if geo_artefact_collector is not None:
            gip = GeometrieInheritanceProcessor(classes=oslo_collector.classes,
                                                geometrie_types=self.geoACollector.geometrie_types,
                                                inheritances=self.oslo_collector.inheritances)
            self.geometry_types = gip.process_inheritances()

    def create_blocks_to_write_from_classes(self, oslo_class: OSLOClass, model_location='',
                                            valid_uri_and_types: Dict = None) -> [str]:
        if not isinstance(oslo_class, OSLOClass):
            raise ValueError(f"Input is not a OSLOClass")

        if oslo_class.objectUri == '':
            raise ValueError(f"OSLOClass.objectUri is invalid. Value = '{oslo_class.objectUri}'")

        if oslo_class.objectUri == 'http://purl.org/dc/terms/Agent':
            pass
        elif not re.match(pattern="^.+/ns/.+#.+", string=oslo_class.objectUri):
            raise ValueError(f"OSLOClass.objectUri is invalid. Value = '{oslo_class.objectUri}'")

        if oslo_class.name == '':
            raise ValueError(f"OSLOClass.name is invalid. Value = '{oslo_class.name}'")

        return self.create_block_from_class(oslo_class, model_location, valid_uri_and_types=valid_uri_and_types)

    def create_block_from_class(self, oslo_class: OSLOClass, model_location: str = '',
                                valid_uri_and_types: Dict = None) -> [str]:
        if valid_uri_and_types is None:
            valid_uri_and_types = {}
        attributen = self.oslo_collector.find_attributes_by_class(oslo_class)
        inheritances = self.oslo_collector.find_inheritances_by_class(oslo_class)
        list_of_geometry_types = self.get_geometry_types_from_uri(oslo_class.objectUri)

        if oslo_class.objectUri in {'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMObject',
                                    'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Derdenobject'}:
            inheritances.append(
                Inheritance(base_name='OTLAsset', base_uri='', class_name='', class_uri='', deprecated_version=''))
            inheritances.append(
                Inheritance(base_name='RelationInteractor', base_uri='', class_name='', class_uri='',
                            deprecated_version=''))

        elif oslo_class.objectUri == 'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#RelatieObject':
            inheritances.append(
                Inheritance(base_name='DavieRelatieAttributes', base_uri='', class_name='', class_uri='',
                            deprecated_version=''))
            inheritances.append(
                Inheritance(base_name='OTLObject', base_uri='', class_name='', class_uri='', deprecated_version=''))

        elif oslo_class.objectUri in {
                'http://purl.org/dc/terms/Agent',
                'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Toegangsprocedure',
                'https://wegenenverkeer.data.vlaanderen.be/ns/abstracten#AbstracteAanvullendeGeometrie'}:
            inheritances.append(
                Inheritance(base_name='OTLObject', base_uri='', class_name='', class_uri='', deprecated_version=''))
            inheritances.append(
                Inheritance(base_name='RelationInteractor', base_uri='', class_name='', class_uri='',
                            deprecated_version=''))

        datablock = ['# coding=utf-8']
        if len(attributen) > 0:
            datablock.append('from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLAttribuut')

        if oslo_class.abstract == 1:
            if len(inheritances) + len(list_of_geometry_types) < 1:
                datablock.append('from abc import abstractmethod, ABC')
            else:
                datablock.append('from abc import abstractmethod')

        if len(inheritances) > 0:
            for inheritance in inheritances:
                if inheritance.base_name in ['OTLAsset', 'OTLObject', 'RelationInteractor', 'DavieRelatieAttributes']:
                    datablock.append(
                        f'from otlmow_model.OtlmowModel.BaseClasses.{inheritance.base_name} import {inheritance.base_name}')
                else:
                    class_directory = 'Classes'
                    ns = None
                    if inheritance.base_uri != 'http://purl.org/dc/terms/Agent':
                        ns, name = get_ns_and_name_from_uri(inheritance.base_uri)
                    if ns is not None:
                        class_directory = get_class_directory_from_ns(ns).replace('/', '.')

                    if inheritance.class_name == 'Agent' and inheritance.base_name in {'AIMDBStatus', 'AIMVersie'}:
                        datablock.append(f'from ..{class_directory}.{inheritance.base_name} '
                                         f'import {inheritance.base_name}')
                    else:
                        datablock.append(f'from ...{class_directory}.{inheritance.base_name} '
                                     f'import {inheritance.base_name}')

        if any(atr.readonly == 1 for atr in attributen):
            raise NotImplementedError("readonly property is assumed to be 0 on value fields")

        list_of_fields = get_fields_to_import_from_list_of_attributes(oslo_collector=self.oslo_collector,
                                                                      attributen=attributen,
                                                                      valid_uri_and_types=valid_uri_and_types)
        base_fields = ['BooleanField', 'ComplexField', 'DateField', 'DateTimeField', 'FloatOrDecimalField',
                       'IntegerField', 'KeuzelijstField', 'UnionTypeField', 'URIField', 'LiteralField',
                       'NonNegIntegerField', 'TimeField', 'StringField', 'UnionWaarden']
        for type_field in list_of_fields:
            if type_field not in base_fields:
                if oslo_class.objectUri == 'http://purl.org/dc/terms/Agent':
                    datablock.append(f'from ..Datatypes.{type_field} import {type_field}')
                else:
                    datablock.append(f'from ...Datatypes.{type_field} import {type_field}')
            else:
                datablock.append(f'from otlmow_model.OtlmowModel.BaseClasses.{type_field} import {type_field}')

        if 'Bevestiging' in oslo_class.objectUri:
            pass

        for geometry_type in list_of_geometry_types:
            datablock.append(f'from otlmow_model.OtlmowModel.GeometrieTypes.{geometry_type} import {geometry_type}')

        datablock.append('')
        datablock.append('')
        datablock.append(f'# Generated with {self.__class__.__name__}. To modify: extend, do not edit')

        inheritances = self.sort_inheritances_for_naampad_naam_interaction(inheritances)
        datablock.append(self.get_class_line_from_class_and_inheritances(oslo_class=oslo_class,
                                                                         inheritances=inheritances,
                                                                         geometry_types=list_of_geometry_types))
        datablock.append(f'    """{oslo_class.definition}"""')
        datablock.append('')
        datablock.append(f"    typeURI = '{oslo_class.objectUri}'")
        datablock.append('    """De URI van het object volgens https://www.w3.org/2001/XMLSchema#anyURI."""')

        if oslo_class.deprecated_version != '':
            oslo_class.deprecated_version = oslo_class.deprecated_version.split('-')[0]
            datablock.append('')
            datablock.append(f"    deprecated_version = '{oslo_class.deprecated_version}'")

        datablock.append('')
        if oslo_class.abstract == 1:
            datablock.append('    @abstractmethod')
        datablock.append('    def __init__(self):')
        datablock.append('        super().__init__()')
        datablock.append('')

        self.add_relations_to_datablock(datablock, oslo_class.objectUri)

        add_attributen_to_data_block(attributen, datablock, for_class_use=True, valid_uri_and_types=valid_uri_and_types)
        if len(inheritances) == 0 and len(attributen) == 0:
            datablock.append('        pass')

        if datablock[-1] == '':
            datablock.pop()

        return datablock

    def get_class_line_from_class_and_inheritances(self, oslo_class: OSLOClass, inheritances: [Inheritance],
                                                   geometry_types: [GeometrieType]) -> str:
        if oslo_class.abstract + len(inheritances) + len(geometry_types) < 1:
            raise NotImplementedError(f"{oslo_class.objectUri} class structure not implemented")
        if oslo_class.abstract == 1 and len(inheritances) + len(geometry_types) < 1:
            return f'class {oslo_class.name}(ABC):'
        if len(inheritances) + len(geometry_types) > 0:
            line = f'class {oslo_class.name}('
            for inheritance in inheritances:
                line += inheritance.base_name + ', '
            for geometry_type in geometry_types:
                line += geometry_type + ', '
            line = line[:-2]
            line += '):'
            return line

        raise NotImplementedError(f"{oslo_class.objectUri} class structure not implemented")

    def get_geometry_types_from_uri(self, object_uri: str) -> [str]:
        if len(self.geometry_types) == 0:
            return []

        geom_type = next((g for g in self.geometry_types if g.objectUri == object_uri), None)
        if geom_type is None:
            return []

        geom_types = []
        if geom_type.geen_geometrie == 1:
            geom_types.append('GeenGeometrie')
        if geom_type.punt3D == 1:
            geom_types.append('PuntGeometrie')
        if geom_type.lijn3D == 1:
            geom_types.append('LijnGeometrie')
        if geom_type.polygoon3D == 1:
            geom_types.append('VlakGeometrie')

        return geom_types

    def add_relations_to_datablock(self, datablock: [str], object_uri: str) -> None:
        relations = self.oslo_collector.find_all_relations(object_uri, allow_duplicates=False)
        if len(relations) == 0:
            return

        for relation in relations:
            deprecated = ''
            if relation.deprecated_version != '':
                relation.deprecated_version = relation.deprecated_version.split('-')[0]
                deprecated = f", deprecated='{relation.deprecated_version}'"
            if relation.richting == 'Unspecified':
                direction = 'u' # u = unidirectional
                direction_comment = '# u = unidirectional'
                if relation.bron_uri == object_uri:
                    target_uri = relation.doel_uri
                else:
                    target_uri = relation.bron_uri
            elif relation.bron_uri == object_uri:
                direction = 'o' # o = direction: outgoing
                direction_comment = '# o = direction: outgoing'
                target_uri = relation.doel_uri
            else:
                direction = 'i' # i = direction: incoming
                direction_comment = '# i = direction: incoming'
                target_uri = relation.bron_uri

            datablock.append(
                f"        self.add_valid_relation(relation='{relation.objectUri}', target='{target_uri}', "
                f"direction='{direction}'{deprecated})  {direction_comment}")
        datablock.append('')

    relation_interactor_uris = {
        'https://wegenenverkeer.data.vlaanderen.be/ns/abstracten#AbstracteAanvullendeGeometrie',
        'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMObject',
        'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Derdenobject',
        'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#Toegangsprocedure',
        'http://purl.org/dc/terms/Agent'}

    def search_recursive_inheritance_for_relation_interactor(self, base_uri: str) -> bool:
        inheritances = self.oslo_collector.find_inheritances_by_class_uri(base_uri)
        for inheritance in inheritances:
            if inheritance.base_uri in self.relation_interactor_uris:
                return True
            if self.search_recursive_inheritance_for_relation_interactor(inheritance.base_uri):
                return True
        return False

    @classmethod
    def sort_inheritances_for_naampad_naam_interaction(cls, inheritances: [Inheritance]):
        naampad = next(
            (i for i in inheritances 
             if i.base_uri == 'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#NaampadObject'), None)
        if naampad is None:
            return inheritances
        naam_inh = next(
            (i for i in inheritances
             if i.base_uri == 'https://wegenenverkeer.data.vlaanderen.be/ns/implementatieelement#AIMNaamObject'), None)
        if naam_inh is None:
            return inheritances
        naam_index = inheritances.index(naam_inh)
        inheritances.pop(inheritances.index(naampad))
        inheritances.insert(naam_index, naampad)

        return inheritances



import dataclasses


@dataclasses.dataclass
class OSLOClass:
    label: str = ''
    name: str = ''
    objectUri: str = ''
    definition: str = ''
    usagenote: str = ''
    abstract: int = -1
    deprecated_version: str = ''



import logging
import re

from otlmow_modelbuilder.AbstractDatatypeCreator import AbstractDatatypeCreator
from otlmow_modelbuilder.OSLOCollector import OSLOCollector
from otlmow_modelbuilder.SQLDataClasses.OSLODatatypePrimitive import OSLODatatypePrimitive


class OTLPrimitiveDatatypeCreator(AbstractDatatypeCreator):
    def __init__(self, oslo_collector: OSLOCollector):
        super().__init__(oslo_collector)
        logging.info("Created an instance of OTLPrimitiveDatatypeCreator")

    def create_block_to_write_from_primitive_types(self, oslo_datatype_primitive: OSLODatatypePrimitive,
                                                   primitive_datatype_validation_rules,
                                                   model_location: str = '') -> [str]:
        if not isinstance(oslo_datatype_primitive, OSLODatatypePrimitive):
            raise ValueError(f"Input is not a OSLODatatypePrimitive")

        if oslo_datatype_primitive.objectUri == '':
            raise ValueError(f"OSLODatatypePrimitive.objectUri is invalid. Value = '{oslo_datatype_primitive.objectUri}'")

        if oslo_datatype_primitive.objectUri in primitive_datatype_validation_rules['valid_uri_and_types'].keys():
            pass
        else:
            match_re = False
            for regex in primitive_datatype_validation_rules["valid_regexes"]:
                match_re = re.match(pattern=regex, string=oslo_datatype_primitive.objectUri)
                if match_re:
                    break
            if not match_re:
                raise ValueError(
                    f"OSLODatatypePrimitive.objectUri is invalid. Value = '{oslo_datatype_primitive.objectUri}'")

        if oslo_datatype_primitive.name == '':
            raise ValueError(f"OSLODatatypePrimitive.name is invalid. Value = '{oslo_datatype_primitive.name}'")

        if '#KwantWrd' in oslo_datatype_primitive.objectUri:
            return self.create_block_to_write_from_complex_primitive_or_union_types(
                oslo_datatype=oslo_datatype_primitive,
                type_field='KwantWrd',
                model_location=model_location)
        if '#Dte' in oslo_datatype_primitive.objectUri:
            return self.create_block_to_write_from_complex_primitive_or_union_types(
                oslo_datatype=oslo_datatype_primitive,
                type_field='Primitive',
                model_location=model_location)



from pathlib import Path


class OTLExtraChecker:
    @classmethod
    def modify_for_extra_checks(cls, directory: Path) -> None:
        cls.modify_naamfield(directory)
        cls.modify_naampad_field(directory)

    @classmethod
    def modify_naamfield(cls, directory: Path) -> None:
        naamfield_path = directory / 'Classes/ImplementatieElement/AIMNaamObject.py'
        if not naamfield_path.exists():
            return

        with open(naamfield_path, 'r') as f:
            lines = f.readlines()

        if 'import re' not in lines:
            lines.insert(1, '\n')
            lines.insert(1, 'import re\n')

        namefield_line_number = lines.index("        self._naam = OTLAttribuut(field=StringField,\n")
        lines[namefield_line_number] = "        self._naam = OTLAttribuut(field=NaamField,\n"

        class_line_number = lines.index('class AIMNaamObject(AIMObject):\n') - 2
        block = ['\n', "class NaamField(StringField):\n",
"    def __init__(self, naam: str, label: str, objectUri: str, definition: str, owner):\n",
"        super().__init__(naam, label, objectUri, definition, owner)\n",
"\n",
"    @classmethod\n",
"    def validate(cls, value, attribuut) -> bool:\n",
'        if not StringField.validate(value, attribuut):\n',
'            return False\n',
"        if re.match(r'^[\\w.\\-]*$', value) is None:\n",
'            return False\n',
"        if hasattr(attribuut.owner, 'naampad') and attribuut.owner.naampad is not None:\n",
"            return attribuut.owner.naampad.split('/')[-1] == value\n",
'        return True\n',
 "\n",
"    @classmethod\n",
"    def create_dummy_data(cls) -> str:\n",
"        return 'dummy'", '\n']

        for index, line in enumerate(block):
            lines.insert(class_line_number + index, line)

        with open(naamfield_path, 'w') as f:
            f.writelines(lines)

    @classmethod
    def modify_naampad_field(cls, directory: Path) -> None:
        naampad_path = directory / 'Classes/ImplementatieElement/NaampadObject.py'
        if not naampad_path.exists():
            return

        with open(naampad_path) as f:
            lines = f.readlines()

        if 'import re' not in lines:
            lines.insert(1, '\n')
            lines.insert(1, 'import re\n')

        lines.remove("from otlmow_model.OtlmowModel.BaseClasses.StringField import StringField\n")

        namefield_line_number = lines.index("        self._naampad = OTLAttribuut(field=StringField,\n")
        lines[namefield_line_number] = "        self._naampad = OTLAttribuut(field=NaampadField,\n"

        namefield_line_number = lines.index("from ...Classes.ImplementatieElement.AIMNaamObject import AIMNaamObject\n")
        lines[namefield_line_number] = "from ...Classes.ImplementatieElement.AIMNaamObject import AIMNaamObject, NaamField\n"

        class_line_number = lines.index('class NaampadObject(AIMNaamObject):\n') - 2
        block = ['\n', "class NaampadField(NaamField):\n",
"    def __init__(self, naam: str, label: str, objectUri: str, definition: str, owner):\n",
"        super().__init__(naam, label, objectUri, definition, owner)\n",
"\n",
"    @classmethod\n",
"    def validate(cls, value, attribuut) -> bool:\n",
"        if re.match(r'^[\\w.\\-]+[/[\\w.\\-]+]*$', value) is None:\n",
"            return False\n",
"        if attribuut.owner.naam is not None:\n",
"            return value.split('/')[-1] == attribuut.owner.naam\n",
"        return True\n",
"\n",
"    @classmethod\n",
"    def create_dummy_data(cls) -> str:\n",
"        return 'dummy/dummy'", '\n']

        for index, line in enumerate(block):
            lines.insert(class_line_number + index, line)

        with open(naampad_path, 'w') as f:
            f.writelines(lines)
